from alaas.server.util import ConfigManager

example_config_path = '/Users/huangyz0918/desktop/zeef/service/example/resnet_triton.yml'


class Server:
    def __int__(self, config_path):
        self.cfg_manager = ConfigManager(config_path)

    def start(self):
        pass



# Copyright (c) 2020, NVIDIA CORPORATION. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Author: Li Yuanming
Email: yuanmingleee@gmail.com
Date: Apr 26, 2022

Example adapted from https://github.com/triton-inference-server/python_backend/blob/main/examples/pytorch/model.py
"""

import torch
import ssl
from torch.utils.dlpack import from_dlpack, to_dlpack
from torchvision.models import resnet18

# FIXME: Seems like PyTorch SSL certificate expires, this is an temporary work around
ssl._create_default_https_context = ssl._create_unverified_context

# triton_python_backend_utils is available in every Triton Python model. You
# need to use this module to create inference requests and responses. It also
# contains some utility functions for extracting information from model_config
# and converting Triton input/output types to numpy types.
import triton_python_backend_utils as pb_utils


class TritonPythonModel:
    """Your Python model must use the same class name. Every Python model
    that is created must have "TritonPythonModel" as the class name.
    """
    __slots__ = ('device', 'resnet_model')

    def initialize(self, args):
        """`initialize` is called only once when the model is being loaded.
        Implementing `initialize` function is optional. This function allows
        the model to intialize any state associated with this model.
        Parameters
        ----------
        args : dict
          Both keys and values are strings. The dictionary keys and values are:
          * model_config: A JSON string containing the model configuration
          * model_instance_kind: A string containing model instance kind
          * model_instance_device_id: A string containing model instance device ID
          * model_repository: Model repository path
          * model_version: Model version
          * model_name: Model name
        """

        # Hard-code device name for PyTorch device placement
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # Instantiate the PyTorch model
        self.resnet_model = resnet18(pretrained=True).to(self.device)

    def execute(self, requests):
        """`execute` must be implemented in every Python model. `execute`
        function receives a list of pb_utils.InferenceRequest as the only
        argument. This function is called when an inference is requested
        for this model. Depending on the batching configuration (e.g. Dynamic
        Batching) used, `requests` may contain multiple requests. Every
        Python model, must create one pb_utils.InferenceResponse for every
        pb_utils.InferenceRequest in `requests`. If there is an error, you can
        set the error argument when creating a pb_utils.InferenceResponse.
        Parameters
        ----------
        requests : list
          A list of pb_utils.InferenceRequest
        Returns
        -------
        list
          A list of pb_utils.InferenceResponse. The length of this list must
          be the same as `requests`
        """

        responses = []

        # Every Python backend must iterate over everyone of the requests
        # and create a pb_utils.InferenceResponse for each of them.
        for request in requests:
            # Get INPUT0
            in_0 = pb_utils.get_input_tensor_by_name(request, 'INPUT0')
            # `from_dlpack` and `to_dlpack` enables zero-copy transfer between
            # PyTorch tensors and Triton output tensors
            in_pytorch_0 = from_dlpack(in_0.to_dlpack())

            out_0 = self.resnet_model(in_pytorch_0.to(self.device))

            # Create output tensors. You need pb_utils.Tensor
            # objects to create pb_utils.InferenceResponse.
            # Same zero-copy transfer enabled by `from_dlpack` and `to_dlpack`.
            out_tensor_0 = pb_utils.Tensor.from_dlpack("OUTPUT0", to_dlpack(out_0))

            # Create InferenceResponse. You can set an error here in case
            # there was a problem with handling this inference request.
            # Below is an example of how you can set errors in inference
            # response:
            #
            # pb_utils.InferenceResponse(
            #    output_tensors=..., TritonError("An error occured"))
            inference_response = pb_utils.InferenceResponse(
                output_tensors=(out_tensor_0,),
            )
            responses.append(inference_response)

        # You should return a list of pb_utils.InferenceResponse. Length
        # of this list must match the length of `requests` list.
        return responses

    def finalize(self):
        """`finalize` is called only once when the model is being unloaded.
        Implementing `finalize` function is optional. This function allows
        the model to perform any necessary clean ups before exit.
        """
        print('Cleaning up...')



"""
Database Manager for Experiment Records.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 17/08/2020
"""

import time

import docker
import pymongo


class DBManager:
    """
    DBManager: A manager for the database (mongoDB) managements.
    """

    def __init__(self, db_name, container_name, ports):
        """
        Parameters.
        :param db_name: the database name.
        :param container_name: the docker container of mongo's name.
        :param ports: the ports mapping.
        """
        self.db_name = db_name
        self.ports = ports
        self.container_name = container_name
        self.docker_client = docker.from_env()
        self.db_container = None

    def start_db_container(self, timeout):
        """
        Start the Docker container.
        :param timeout: the time to wait the docker started.
        :return: None
        """
        self.db_container = self.docker_client.containers.run(
            image='mongo', remove=True, detach=True, ports=self.ports,
            name=self.container_name
        )
        time.sleep(timeout)

    def stop_db_container(self):
        """
        Stop the database container.
        :return: None
        """
        self.db_container.stop()

    def get_container_status(self):
        """
        check the docker container's status, could be 'running' or 'exited'.
        :return: None
        """
        return self.db_container.status

    @staticmethod
    def get_mongo_client(address='mongodb://localhost:27017/'):
        """
        Get the mongo Python client
        :param address: the database service address, contains ports and host.
        :return: client.
        """
        return pymongo.MongoClient(address)

    @staticmethod
    def create_db(name='bmk', address='mongodb://localhost:27017/', username='hyz', password='123456'):
        """
        Create the database to store the benchmarking results.
        :param name: the database's name.
        :param address: the database service address, contains ports and host.
        :param username: the database user name.
        :param password: the database password.
        :return: a new table.
        """
        mongo_client = pymongo.MongoClient(address,
                                           username=username,
                                           password=password,
                                           authSource=name,
                                           authMechanism='SCRAM-SHA-256')
        return mongo_client[name]

    @staticmethod
    def create_db_without_auth(name='pycontinual', address='mongodb://localhost:27017/'):
        """
        Create the database to store the benchmarking results.
        :param name: the database's name.
        :param address: the database service address, contains ports and host.
        :return: a new table.
        """
        mongo_client = pymongo.MongoClient(address,
                                           authSource=name,
                                           authMechanism='SCRAM-SHA-256')
        return mongo_client[name]

    @staticmethod
    def create_doc(database, doc_name='experiment'):
        return database[doc_name]

    @staticmethod
    def insert_record(db_doc, record: dict):
        """
        Insert single record into the database.
        :param db_doc: the document of the created database.
        :param record: the single record needs to be inserted.
        :return:
        """
        return db_doc.insert_one(record).inserted_id

    @staticmethod
    def insert_record_list(db_doc, record_list: list):
        """
        Insert records into the database.
        :param db_doc: the document of the created database.
        :param record_list: the records need to be inserted.
        :return:
        """
        return db_doc.insert_many(record_list).inserted_ids

    @staticmethod
    def delete_one_record(db_doc, query):
        """
        Query the records from the database table, and delete it.
        :param db_doc: the document of database.
        :param query: the query object (e.g., {"address": "Park Lane 38"}).
        :return: the deleted record id.
        """
        return db_doc.delete_one(query)

    @staticmethod
    def delete_records(db_doc, query):
        """
        Query the records from the database table, and delete them.
        :param db_doc: the document of database.
        :param query: the query object (e.g., {"address": "Park Lane 38"}).
        :return: the deleted records id.
        """
        return db_doc.delete_many(query)

    @staticmethod
    def delete_db_set(db_doc):
        """
        Delete the specific database table.
        :param db_doc: the documents of database you want to delete.
        """
        return db_doc.drop()

    @staticmethod
    def delete_db(client: pymongo.MongoClient, db_name: str):
        """
        Delete the specific database table.
        :param db_name: the database you want to delete.
        :param client: the mongo client.
        """
        client.drop_database(db_name)

    @staticmethod
    def update_one_record(db_doc, query, new_values):
        """
        Update the single record.
        :param db_doc: the document of database.
        :param query: the query object (e.g., {"address": "Park Lane 38"}).
        :param new_values: the new value objects (e.g., {"$set":{"address": "Canyon 123"}}).
        :return: Updated record's id.
        """
        return db_doc.update_one(query, new_values)

    @staticmethod
    def update_records(db_doc, query, new_values):
        """
        Update the single record.
        :param db_doc: the document of database.
        :param query: the query object (e.g., {"address": "Park Lane 38"}).
        :param new_values: the new value objects (e.g., {"$set":{"address": "Canyon 123"}}).
        :return: Updated record's id.
        """
        return db_doc.update_many(query, new_values)

    @staticmethod
    def query_record(db_doc, query):
        """
        Query the records from the database table.
        :param db_doc: the document of database.
        :param query: the query object (e.g., {"address": "Park Lane 38"}).
        :return: the matched records.
        """
        return db_doc.find(query)



import io
import torch
import torch.jit
import numpy as np
from PIL import Image
from torchvision import transforms
import torch.nn.functional as F

from .base import ServeEngine


class TorchServe(ServeEngine):
    def __init__(self, model_path):
        # set device
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        # load the latest version of a TorchScript model
        self.model = torch.jit.load(model_path, map_location=self.device)
        self.model.eval()

    def batch_predict(self, inputs: torch.Tensor, is_prob=False):
        inputs = inputs.to(self.device)
        outputs = self.model(inputs)
        if is_prob:
            outputs = F.softmax(outputs, dim=1)

        if isinstance(outputs, tuple):
            return outputs[0].cpu().detach().numpy()
        else:
            return outputs.cpu().detach().numpy()

    # TODO: only for image model tests.
    @staticmethod
    def transform_image(image_bytes):
        t = transforms.Compose([transforms.Resize(255),
                                transforms.CenterCrop(224),
                                transforms.ToTensor(),
                                transforms.Normalize(
                                    [0.485, 0.456, 0.406],
                                    [0.229, 0.224, 0.225])])
        image = Image.open(io.BytesIO(image_bytes))
        return t(image).unsqueeze(0)


if __name__ == '__main__':
    # TODO: only for image model tests.
    resnet_50_path = "/Users/huangyz0918/Downloads/torchscript/1.zip"
    example_image = "/Users/huangyz0918/desktop/cat.jpeg"
    with open(example_image, 'rb') as f:
        image_bytes = f.read()
        tensor = TorchServe.transform_image(image_bytes=image_bytes)
        server = TorchServe(resnet_50_path)
        prediction = server.batch_predict(tensor, is_prob=True)
        print("top, top_class: ", np.sort(prediction)[0][-1])
        print(np.shape(prediction))



import numpy as np
from .base import Strategy
from alaas.server.serving import ServeClient
from alaas.server.util import chunks


class LeastConfidence(Strategy):
    """
    Least Confidence Sampling.
    Reference: A Sequential Algorithm for Training Text Classifiers (1994).
    (https://arxiv.org/pdf/cmp-lg/9407020)
    """

    def __init__(self, infer_func, source_data):
        super(LeastConfidence, self).__init__(infer_func, source_data)

    def query(self, n):
        probs = self.infer_func(self.source_data, is_prob=True)
        uncertainties = np.amax(probs, axis=1)
        return uncertainties.argsort()[:n]


class LeastConfidenceTriton(Strategy):
    """
    Least Confidence Sampling using Triton Inference Server.
    Reference: A Sequential Algorithm for Training Text Classifiers (1994).
    (https://arxiv.org/pdf/cmp-lg/9407020)
    """

    def __init__(self, source_data, model_name, batch_size, address='localhost:8900'):
        super(LeastConfidenceTriton, self).__init__(None, source_data)
        self.model_name = model_name
        self.batch_size = batch_size
        self.address = address

    def query(self, n):
        probs = []
        triton_client = ServeClient()
        for batch in chunks(self.source_data, self.batch_size):
            for item in triton_client.infer(batch, self.model_name, address=self.address):
                probs.append(item)
        uncertainties = np.amax(probs, axis=1)
        return uncertainties.argsort()[:n]



from torch.utils.data import DataLoader

import torchvision
import torchvision.transforms as transforms

from alaas.server.serving import TorchServe
from alaas.server.strategy import LeastConfidence, LeastConfidenceTriton
from alaas.server.util import ConfigManager


def prepare_data(batch_size):
    """
    Data pre-processing function for Resnet.
    @param batch_size: the input batch size for data loader (not for active learning).
    @return: torch data loader.
    """
    transform = transforms.Compose([
        transforms.ToTensor(),
        transforms.Resize(256),
        transforms.CenterCrop(224),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])

    dataset = torchvision.datasets.CIFAR10(root='./data', train=False, download=True, transform=transform)
    return DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=2)


def triton_example(model_name, budget, batch_size, address='localhost:8900'):
    # print("preparing data...")
    inputs, targets = next(iter(prepare_data(100)))
    # print(f"start active learning, query number: {budget}...")
    strategy = LeastConfidenceTriton(source_data=inputs.numpy(), model_name=model_name, batch_size=batch_size,
                                     address=address)
    data_index_list = strategy.query(budget)
    print("selected data index: ", data_index_list)


def torch_example(budget, batch_size=10000):
    resnet_50_path = "/Users/huangyz0918/Downloads/torchscript/1.zip"
    server = TorchServe(resnet_50_path)

    for batch_idx, (inputs, targets) in enumerate(prepare_data(batch_size)):
        strategy = LeastConfidence(server.batch_predict, source_data=inputs)
        data_index_list = strategy.query(budget)
        print("selected data index: ", data_index_list)


if __name__ == '__main__':
    # read the global configuration.
    config_path = 'resnet_triton.yml'
    cfg_manager = ConfigManager(config_path)
    strategy = cfg_manager.get_al_config()['strategy']
    algorithm = strategy['algorithm']
    batch_size = strategy['infer_model']['batch_size']
    model_name = strategy['infer_model']['name']
    budget = cfg_manager.get_al_config()['budget']
    address = cfg_manager.get_al_config()['al_server']['address']

    # run the built-in pytorch server example.
    # torch_example(cfg_manager.get_al_config()['budget'])
    # run th NVIDIA triton inference server example.
    print("=====================")
    print(f"active learning: {algorithm}, budget: {budget}")
    print(f"AL model: {model_name}, inference batch: {batch_size}")
    print("=====================")
    triton_example(model_name=model_name, batch_size=batch_size, budget=budget, address=address)



"""
Configuration file manager for active learning as a service.
@author: huangyz0918 (huangyz0918@gmail.com)
@date: 01/05/2022
"""

import yaml


class ConfigManager:

    def __init__(self, config_path):
        self.config_dict = ConfigManager.load_config(config_path)

    @property
    def dry_run(self):
        return self.config_dict.get('dry_run', False)

    @staticmethod
    def load_config(config_path):
        result_dict = None
        with open(config_path, 'r') as stream:
            try:
                result_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
        return result_dict

    def get_job_name(self):
        return self.config_dict['name']

    def get_job_version(self):
        return self.config_dict['version']

    def get_al_config(self):
        return self.config_dict['active_learning']



class Strategy:
    def __init__(self, infer_func, source_data):
        self.infer_func = infer_func
        self.source_data = source_data

    def query(self, number):
        """
        The query function that calls the specific active learning strategy.
        :param number: the query number.
        :return: the data ids in a numpy array.
        """
        pass



import numpy as np

import tritonclient.http as httpclient
from tritonclient.utils import np_to_triton_dtype


class ServeEngine:
    """
    ServeEngine: Pytorch serving engine built-in.
    """

    def batch_predict(self, inputs, is_prob=False):
        pass


class ServeClient:
    """
    ServeClient: for inference service provided by Triton Inference Server (NVIDIA).
    """

    @staticmethod
    def infer(inputs_data, model_name, address="localhost:8900"):
        with httpclient.InferenceServerClient(address) as cli:
            inputs = [httpclient.InferInput("INPUT0", inputs_data.shape, np_to_triton_dtype(inputs_data.dtype))]
            inputs[0].set_data_from_numpy(inputs_data)

            outputs = [httpclient.InferRequestedOutput("OUTPUT0")]

            response = cli.infer(
                model_name,
                inputs,
                request_id=str(1),
                outputs=outputs
            )

            outputs_data = response.as_numpy("OUTPUT0")
            return outputs_data


if __name__ == '__main__':
    shape = (4, 3, 224, 224)
    input_data = np.random.rand(*shape).astype(np.float32)
    client = ServeClient()
    outputs = client.infer(input_data, 'resnet')
    print(np.shape(outputs))
